#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Load base config and export package constants.

The ``foremast`` configuration file is read from the following locations in
descending order. First found wins.

* ``./.foremast/foremast.cfg``
* ``~/.foremast/foremast.cfg``
* ``/etc/foremast/foremast.cfg``

.. literalinclude:: ../../src/foremast/templates/configs/foremast.cfg.example
   :language: ini
"""
import ast
import importlib.util
import json
import logging
import sys
from configparser import ConfigParser
from os import getcwd, getenv, path
from os.path import exists, expanduser, expandvars

LOG = logging.getLogger(__name__)
LOGGING_FORMAT = '%(asctime)s [%(levelname)s] %(name)s:%(funcName)s:%(lineno)d - %(message)s'
SHORT_LOGGING_FORMAT = '[%(levelname)s] %(message)s'

logging.basicConfig(format=LOGGING_FORMAT)
logging.getLogger(__package__.split('.')[0]).setLevel(logging.INFO)

GOOD_STATUSES = frozenset(('SUCCEEDED',))
SKIP_STATUSES = frozenset(('NOT_STARTED',))

DEFAULT_DYNAMIC_CONFIG_FILE = '{path}/config.py'.format(path=getcwd())
"""Default `config.py` file path is in the current directory.

To override, use the `FOREMAST_CONFIG_FILE` environment variable.
"""


def validate_key_values(config_handle, section, key, default=None):
    """Warn when *key* is missing from configuration *section*.

    Args:
        config_handle (configparser.ConfigParser): Instance of configurations.
        section (str): Name of configuration section to retrieve.
        key (str): Configuration key to look up.
        default (object): Default object to use when *key* is not found.

    Returns:
        object: ``str`` when *key* exists, otherwise *default* object.
    """
    if section not in config_handle:
        LOG.info('Section missing from configurations: [%s]', section)

    try:
        value = config_handle[section][key]
    except KeyError:
        LOG.warning('[%s] missing key "%s", using %r.', section, key, default)
        value = default

    return value


def extract_formats(config_handle):
    """Get application formats.

    See :class:`gogoutils.Formats` for available options.

    Args:
        config_handle (configparser.ConfigParser): Instance of configurations.

    Returns:
        dict: Formats in ``{$format_type: $format_pattern}``.

    """
    configurations = dict(config_handle)
    formats = dict(configurations.get('formats', {}))
    return formats


def load_dynamic_config(config_file=DEFAULT_DYNAMIC_CONFIG_FILE):
    """Load and parse dynamic config"""
    dynamic_configurations = {}

    # Insert config path so we can import it
    sys.path.insert(0, path.dirname(path.abspath(config_file)))
    try:
        config_file_spec = importlib.util.spec_from_file_location("config", config_file)
        config_file_module = importlib.util.module_from_spec(config_file_spec)
        config_file_spec.loader.exec_module(config_file_module)
        dynamic_configurations = config_file_module.CONFIG
    except ImportError:
        # Provide a default if config not found
        LOG.error('ImportError: Unable to load dynamic config. Check config.py file imports!')
    except AttributeError:
        LOG.error('AttributeError: Unable to load CONFIG from %s', config_file)

    return dynamic_configurations


def find_config():
    """Look for **foremast.cfg** in config_locations or ``./config.py``.

    Raises:
        SystemExit: No configuration file found.

    Returns:
        dict: Found dynamic or static configuration.

    """
    config_locations = [
        '/etc/foremast/foremast.cfg',
        expanduser('~/.foremast/foremast.cfg'),
        './.foremast/foremast.cfg',
    ]
    configurations = ConfigParser()

    cfg_file = configurations.read(config_locations)
    dynamic_config_file = getenv('FOREMAST_CONFIG_FILE', DEFAULT_DYNAMIC_CONFIG_FILE)

    if cfg_file:
        LOG.info('Loading static configuration file.')
    elif exists(dynamic_config_file):
        LOG.info('Loading dynamic configuration file.')
        configurations = load_dynamic_config(config_file=dynamic_config_file)
    else:
        config_locations.append(dynamic_config_file)
        LOG.warning('No configuration found in the following locations:\n%s', '\n'.join(config_locations))
        LOG.warning('Using defaults...')

    return dict(configurations)


def _remove_empty_entries(entries):
    """Remove empty entries in a list"""
    valid_entries = []
    for entry in set(entries):
        if entry:
            valid_entries.append(entry)
    return sorted(valid_entries)


def _convert_string_to_native(value):
    """Convert a string to its native python type"""
    result = None

    try:
        result = ast.literal_eval(str(value))
    except (SyntaxError, ValueError):
        # Likely a string
        result = value.split(',')
    return result


def _generate_security_groups(config_key):
    """Read config file and generate security group dict by environment.

    Args:
        config_key (str): Configuration file key

    Returns:
        dict: of environments in {'env1': ['group1', 'group2']} format
    """
    raw_default_groups = validate_key_values(CONFIG, 'base', config_key, default='')
    default_groups = _convert_string_to_native(raw_default_groups)
    LOG.debug('Default security group for %s is %s', config_key, default_groups)

    entries = {}
    for env in ENVS:
        entries[env] = []

    if isinstance(default_groups, (list)):
        groups = _remove_empty_entries(default_groups)
        for env in entries:
            entries[env] = groups
    elif isinstance(default_groups, (dict)):
        entries.update(default_groups)

    LOG.debug('Generated security group: %s', entries)
    return entries


CONFIG = find_config()

API_URL = validate_key_values(CONFIG, 'base', 'gate_api_url')
GIT_URL = validate_key_values(CONFIG, 'base', 'git_url')
DOMAIN = validate_key_values(CONFIG, 'base', 'domain', default='example.com')
ENVS = set(validate_key_values(CONFIG, 'base', 'envs', default='').split(','))
ENV_CONFIGS = validate_key_values(CONFIG, 'base', 'env_configs', default={})
VPC_NAME = validate_key_values(CONFIG, 'base', 'vpc_name', default='vpc')
REGIONS = set(validate_key_values(CONFIG, 'base', 'regions', default='').split(','))
RUNWAY_BASE_PATH = validate_key_values(CONFIG, 'base', 'runway_base_path', default='runway')
TEMPLATES_PATH = validate_key_values(CONFIG, 'base', 'templates_path')
AMI_JSON_URL = validate_key_values(CONFIG, 'base', 'ami_json_url')
DEFAULT_RUN_AS_USER = validate_key_values(CONFIG, 'base', 'default_run_as_user', default=None)
DEFAULT_SECURITYGROUP_RULES = _generate_security_groups('default_securitygroup_rules')
DEFAULT_EC2_SECURITYGROUPS = _generate_security_groups('default_ec2_securitygroups')
DEFAULT_ELB_SECURITYGROUPS = _generate_security_groups('default_elb_securitygroups')

EC2_PIPELINE_TYPES = tuple(validate_key_values(CONFIG, 'base', 'ec2_pipeline_types', default='ec2,rolling').split(','))
"""Comma separated list of Pipeline Types to treat as EC2 deployments.

This is useful when defining custom Pipeline Types. When Pipeline Type matches,
EC2 specific data is used in deployment, such as Auto Scaling Groups and
Availability Zones.

    | *Default*: ``ec2,rolling``
    | *Required*: No
    | *Example*: ``ec2,infrastructure,propeller``
"""

SECURITYGROUP_REPLACEMENTS = _convert_string_to_native(
    validate_key_values(
        CONFIG,
        'base',
        'securitygroup_replacements',
        default='{}',
    ))
GITLAB_TOKEN = validate_key_values(CONFIG, 'credentials', 'gitlab_token')
SLACK_TOKEN = validate_key_values(CONFIG, 'credentials', 'slack_token')
GATE_AUTHENTICATION = validate_key_values(CONFIG, 'credentials', 'gate_authentication', default={})
DEFAULT_TASK_TIMEOUT = validate_key_values(CONFIG, 'task_timeouts', 'default', default=120)
TASK_TIMEOUTS = json.loads(validate_key_values(CONFIG, 'task_timeouts', 'envs', default="{}"))
ASG_WHITELIST = set(validate_key_values(CONFIG, 'whitelists', 'asg_whitelist', default='').split(','))
APP_FORMATS = extract_formats(CONFIG)
GATE_CLIENT_CERT = expandvars(expanduser(validate_key_values(CONFIG, 'base', 'gate_client_cert', default='')))
GATE_CA_BUNDLE = expandvars(expanduser(validate_key_values(CONFIG, 'base', 'gate_ca_bundle', default='')))
LINKS = _convert_string_to_native(validate_key_values(CONFIG, 'links', 'default', default='{}'))

GCP_ENVS = validate_key_values(CONFIG, 'gcp', 'envs', default={})

HEADERS = {
    'accept': '*/*',
    'content-type': 'application/json',
    'user-agent': 'foremast',
}

# Define which pipeline types are permitted and which cloud they map to
AWS_MANUAL_TYPES = set(filter(None, set(validate_key_values(CONFIG, 'base', 'aws_manual_types', default='manual')
                                        .split(','))))
AWS_TYPES = set(
    validate_key_values(CONFIG, 'base', 'aws_types', default='ec2,lambda,s3,datapipeline,stepfunction,rolling')
    .split(',')).union(AWS_MANUAL_TYPES)
GCP_MANUAL_TYPES = set(filter(None, set(validate_key_values(CONFIG, 'base', 'gcp_manual_types', default='')
                                        .split(','))))
GCP_TYPES = set(validate_key_values(CONFIG, 'base', 'gcp_types', default='cloudfunction')
                .split(',')).union(GCP_MANUAL_TYPES)

# Previous to Foremast 5.x, only AWS was supported and only MANUAL_TYPES and ALLOWED_TYPES existed
# to support old aws-only configs from <5.x, check if the old config structure is used
# and move the old config variables to the new names accordingly
_legacy_manual_types = validate_key_values(CONFIG, 'base', 'manual_types', default=None)
if _legacy_manual_types is not None:
    LOG.warning("manual_types is deprecated, use aws_manual_types and gcp_manual_types instead")
    AWS_MANUAL_TYPES = set(_legacy_manual_types.split(','))

_legacy_allowed_types = validate_key_values(CONFIG, 'base', 'types', default=None)
if _legacy_allowed_types is not None:
    LOG.warning("types is deprecated, use gcp_types and aws_types instead")
    AWS_TYPES = set(_legacy_allowed_types.split(','))

# Allowed and manual pipelines types for all cloud providers
ALLOWED_TYPES = AWS_TYPES.union(GCP_TYPES)
MANUAL_TYPES = AWS_MANUAL_TYPES.union(GCP_MANUAL_TYPES)

# Check if types have accidentally been defined for both clouds
_duplicate_allowed_types = AWS_TYPES.intersection(GCP_TYPES)
if _duplicate_allowed_types:
    LOG.warning("The following pipeline types are defined as allowed types for multiple cloud providers: '{}' "
                "this can cause unpredictable deployments"
                .format(_duplicate_allowed_types))
_duplicate_manual_types = AWS_MANUAL_TYPES.intersection(GCP_MANUAL_TYPES)
if _duplicate_manual_types:
    LOG.warning("The following pipeline types are defined as manual types for multiple cloud providers: '{}' "
                "this can cause unpredictable deployments"
                .format(_duplicate_manual_types))



#   Foremast - Pipeline Tooling
#
#   Copyright 2020 Redbox Automated Retail, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""GCP IAM Policy management"""

import logging

import googleapiclient.discovery

LOG = logging.getLogger(__name__)


def get_policy(credentials, project_id):
    """Gets IAM policy for a project."""

    service = googleapiclient.discovery.build(
        "cloudresourcemanager", "v1", credentials=credentials,
        cache_discovery=False
    )
    policy = (
        service.projects()
        .getIamPolicy(
            resource=project_id,
            # Request policy version 3 to ensure conditional bindings are included
            # https://cloud.google.com/iam/docs/policies
            body={"options": {"requestedPolicyVersion": 3}},
        )
        .execute()
    )

    return policy


def set_policy(credentials, project_id, policy):
    """Sets IAM policy for a project."""

    minimum_policy_version = 3
    if policy["version"] < minimum_policy_version:
        # The policy may be version 1 as this is the default in GCP
        # it needs to be at least version 3 to support conditional bindings
        LOG.debug("Overriding GCP IAM Policy version, was %s is now %s", policy["version"], minimum_policy_version)
        policy["version"] = minimum_policy_version

    service = googleapiclient.discovery.build(
        "cloudresourcemanager", "v1", credentials=credentials,
        cache_discovery=False
    )

    policy = (
        service.projects()
        .setIamPolicy(resource=project_id, body={"policy": policy})
        .execute()
    )

    LOG.info("Updated IAM Policy for project %s", project_id)

    return policy


def modify_policy_remove_member(policy, member):
    """Removes a member from a role binding if the role is in the given roles.
    Returns true if that member was found and removed, false is the member was
    not found."""
    was_updated = False

    # Policy has no bindings to remove
    if "bindings" not in policy:
        return was_updated

    for binding in policy["bindings"]:
        # If member exists in the binding, remove the member
        if "members" in binding and member in binding["members"]:
            LOG.debug("Removed %s from role %s", member, binding["role"])
            binding["members"].remove(member)
            was_updated = True

    return was_updated


def modify_policy_add_binding(policy, role, member):
    """Adds a new role binding to a policy."""

    binding = {
        "role": role,
        "members": [member]
    }
    policy["bindings"].append(binding)

    LOG.info("Added role binding for %s from role %s", member, role)

    return policy



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Create Cloudwatch events."""
import argparse
import logging

from ...args import add_app, add_debug, add_env, add_region
from ...consts import LOGGING_FORMAT
from .cloudwatch_event import create_cloudwatch_event


def main():
    """Create any Cloudwatch events related Resources."""
    logging.basicConfig(format=LOGGING_FORMAT)

    parser = argparse.ArgumentParser(description=main.__doc__)
    add_debug(parser)
    add_app(parser)
    add_env(parser)
    add_region(parser)
    args = parser.parse_args()

    logging.getLogger(__package__.split('.')[0]).setLevel(args.debug)

    assert create_cloudwatch_event(**vars(args))


if __name__ == '__main__':
    main()



"""Verify AWS Lambda Event creation."""
import copy
from unittest import mock

from foremast.awslambda.awslambdaevent import LambdaEvent

TRIGGERS_MULTIPLE_FILTER = [
    {
        "type": "s3",
        "bucket": "my.shared.bucket",
        "events": [
            "s3:ObjectCreated:*"
        ],
        "prefix": "",
        "suffix": ".json"
    },
    {
        "type": "s3",
        "bucket": "my.shared.bucket",
        "events": [
            "s3:ObjectCreated:*"
        ],
        "prefix": "",
        "suffix": ".tar.gz"
    }
]

TRIGGERS_BUCKET_A = [
    {
        "type": "s3",
        "bucket": "my.shared.bucket",
        "events": [
            "s3:ObjectCreated:*"
        ],
        "prefix": "",
        "suffix": ".json"
    }
]

TRIGGERS_BUCKET_B = [
    {
        "type": "s3",
        "bucket": "my.other.shared.bucket",
        "events": [
            "s3:ObjectCreated:*"
        ],
        "prefix": "",
        "suffix": ".tar.gz"
    }
]

def get_properties_with_triggers(triggers):
    return {
        'app': {
            'lambda_memory': 0,
            'lambda_timeout': 0,
            'lambda_environment': None,
            'lambda_layers': None,
            'lambda_dlq': None,
            'lambda_tracing': None,
            'lambda_destinations': None,
            'lambda_subnet_count': None,
            'lambda_filesystems': None,
            'lambda_provisioned_throughput': None
        },
        "lambda_triggers": triggers
    }

@mock.patch('foremast.awslambda.s3_event.s3_event.add_lambda_permissions')
@mock.patch('foremast.awslambda.s3_event.s3_event.get_lambda_alias_arn')
@mock.patch('foremast.awslambda.s3_event.s3_event.boto3')
@mock.patch('foremast.awslambda.awslambdaevent.remove_all_lambda_permissions')
@mock.patch('foremast.awslambda.awslambdaevent.create_s3_event')
@mock.patch('foremast.awslambda.awslambdaevent.get_properties')
def test_create_s3_event_multiple_filters(mock_get_properties, mock_create_s3_event, mock_remove_perms, mock_boto3, mock_arn, mock_perms):
    """Try to create a lambda with two s3 triggers from the same bucket."""
    triggers = TRIGGERS_MULTIPLE_FILTER
    properties = get_properties_with_triggers(TRIGGERS_MULTIPLE_FILTER)
    mock_get_properties.return_value = properties
    mock_arn.return_value = 'fake_arn'

    events = LambdaEvent(app='test_app', env='test_env', region='us-east-1', prop_path='other')
    events.create_lambda_events()

    mock_create_s3_event.assert_called_with(app_name='test_app', env='test_env', region='us-east-1', bucket='my.shared.bucket', triggers=triggers)

@mock.patch('foremast.awslambda.s3_event.s3_event.add_lambda_permissions')
@mock.patch('foremast.awslambda.s3_event.s3_event.get_lambda_alias_arn')
@mock.patch('foremast.awslambda.s3_event.s3_event.boto3')
@mock.patch('foremast.awslambda.awslambdaevent.remove_all_lambda_permissions')
@mock.patch('foremast.awslambda.awslambdaevent.create_s3_event')
@mock.patch('foremast.awslambda.awslambdaevent.get_properties')
def test_create_s3_event_multiple_buckets(mock_get_properties, mock_create_s3_event, mock_remove_perms, mock_boto3, mock_arn, mock_perms):
    """Try to create a lambda with two S3 triggers from different buckets."""

    triggers = TRIGGERS_BUCKET_A + TRIGGERS_BUCKET_B
    properties = get_properties_with_triggers(triggers)
    mock_get_properties.return_value = properties
    mock_arn.return_value = 'fake_arn'

    events = LambdaEvent(app='test_app', env='test_env', region='us-east-1', prop_path='other')
    events.create_lambda_events()

    s3_calls = [
        mock.call(app_name='test_app', env='test_env', region='us-east-1', bucket='my.shared.bucket', triggers=TRIGGERS_BUCKET_A),
        mock.call(app_name='test_app', env='test_env', region='us-east-1', bucket='my.other.shared.bucket', triggers=TRIGGERS_BUCKET_B)
    ]

    mock_create_s3_event.assert_has_calls(s3_calls, any_order=True)



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from configparser import ConfigParser
from unittest.mock import patch

from foremast.consts import (ALLOWED_TYPES, DEFAULT_SECURITYGROUP_RULES, EC2_PIPELINE_TYPES, RUNWAY_BASE_PATH,
                             _generate_security_groups, extract_formats, validate_key_values)


def test_consts_extract_formats():
    """Test extract_formats()"""

    format_defaults = {
        'domain': 'example.com',
        'app': '{project}',
    }

    config = ConfigParser(defaults=format_defaults)
    config.add_section('formats')

    results = extract_formats(config)
    assert 'example.com' == results['domain']


def test_consts_pipeline_types():
    """Default types should be set."""
    assert 'ec2' in ALLOWED_TYPES
    assert 'lambda' in ALLOWED_TYPES


def test_consts_default_securitygroup_rules():
    """Test this const is a dict"""
    assert isinstance(DEFAULT_SECURITYGROUP_RULES, dict)
    # assert DEFAULT_SECURITYGROUP_RULES == {} or {'': []}


@patch('foremast.consts.validate_key_values')
def test_parse_security_group(values):
    """Parse out security group entries"""
    values.return_value = 'g1'
    assert {'': ['g1']} == _generate_security_groups('default_ec2_securitygroups')

    values.return_value = 'g1,g2'
    assert {'': ['g1', 'g2']} == _generate_security_groups('default_ec2_securitygroups')


@patch('foremast.consts.validate_key_values', return_value="{'': ['g3']}")
def test_parse_security_group_dict(mock_keys):
    """Validate security groups are handled properly if dictionary is in dyanmic config"""
    assert {'': ['g3']} == _generate_security_groups('default_elb_securitygroups')


def test_parse_ec2_pipeline_types():
    """Validate EC2 Pipeline Types."""
    assert EC2_PIPELINE_TYPES == ('ec2', 'rolling')
    assert isinstance(EC2_PIPELINE_TYPES, tuple)

def test_consts_default_runway_base_path():
    """Validate default runway base path remains unchanged"""
    assert RUNWAY_BASE_PATH == 'runway'

def test_consts_custom_runway_base_path():
    """Validate custom runway base path changes as expected"""
    sample_config = {'base': {'runway_base_path': 'custom'}}
    result = validate_key_values(sample_config, 'base', 'runway_base_path', default='runway')
    assert result == 'custom'



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Ensure AMI names can be translated."""
import json
from unittest import mock

import pytest
from foremast.utils import ami_lookup


@mock.patch('foremast.utils.lookups.GITLAB_TOKEN', new=True)
@mock.patch('foremast.utils.lookups._get_ami_file')
def test_ami_lookup(ami_file):
    """AMI lookup should contact GitLab for JSON table and resolve."""
    sample_dict = {
        'base_fedora': 'ami-xxxx',
        'tomcat8': 'ami-yyyy',
    }
    sample_json = json.dumps(sample_dict)
    ami_file.return_value = sample_json
    with pytest.warns(UserWarning):
        assert ami_lookup(name='base_fedora') == 'ami-xxxx'
        assert ami_lookup(region='us-west-2') == 'ami-yyyy'


@mock.patch('foremast.utils.lookups.AMI_JSON_URL', new=True)
@mock.patch('foremast.utils.lookups._get_ami_dict')
def test_dict_lookup(ami_file_dict):
    """AMI lookup using json url."""
    sample_dict = {
        'us-east-1': {
            'base_fedora': 'ami-xxxx',
        },
        'us-west-2': {
            'tomcat8': 'ami-yyyy',
        }
    }
    ami_file_dict.return_value = sample_dict
    assert ami_lookup(region='us-east-1', name='base_fedora') == 'ami-xxxx'
    assert ami_lookup(region='us-west-2', name='tomcat8') == 'ami-yyyy'


def test_no_external_lookup():
    """AMI lookup not using json or gitlab."""
    assert ami_lookup(region='us-east-1', name='no_external') == 'no_external'



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Test validity of json in templates"""
import json

import pytest

from foremast.exceptions import ForemastTemplateNotFound
from foremast.utils import get_template


def test_get_template():
    with pytest.raises(ForemastTemplateNotFound):
        template = get_template(template_file='doesnotexist.json.j2')


def valid_json(template, data):
    parsed_template = get_template(template_file=template, data=data)

    assert type(json.loads(parsed_template)) == dict


def test_valid_json_configs():

    data = {
        'env': 'dev',
        'profile': 'profile',
        'app': 'testapp',
    }

    valid_json(template='configs/configs.json.j2', data=data)


def test_valid_json_pipeline():
    data = {}
    valid_json(template='configs/pipeline.json.j2', data=data)



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Handle API Gateway events"""

import logging

import boto3
import botocore
from tryagain import retries

from ...utils import (add_lambda_permissions, get_details, get_env_credential, get_lambda_alias_arn,
                      get_lambda_arn, get_properties)

LOG = logging.getLogger(__name__)


class APIGateway:
    """Class to handle API Gateway and Lambda integration.

    Args:
        app (str): Application Name
        env (str): Environment/account for deployments
        region (str): AWS Region
        rules (dict): Trigger settings
        prop_path (str): Path to the raw.properties.json
    """

    def __init__(self, app='', env='', region='', rules={}, prop_path=''):
        self.log = logging.getLogger(__name__)
        self.generated = get_details(app=app, env=env)
        self.trigger_settings = rules
        self.app_name = self.generated.app_name()
        self.env = env
        self.account_id = get_env_credential(env=self.env)['accountId']
        self.region = region
        self.properties = get_properties(properties_file=prop_path, env=self.env, region=self.region)

        session = boto3.Session(profile_name=env, region_name=region)
        self.client = session.client('apigateway')
        self.lambda_client = session.client('lambda')
        self.api_version = self.lambda_client.meta.service_model.api_version

        self.api_id = self.find_api_id()
        self.resource_id, self.parent_id = self.find_resource_ids()

    def find_api_id(self):
        """Given API name, find API ID."""
        allapis = self.client.get_rest_apis()
        api_name = self.trigger_settings['api_name']
        api_id = None
        for api in allapis['items']:
            if api['name'] == api_name:
                api_id = api['id']
                self.log.info("Found API for: %s", api_name)
                break
        else:
            api_id = self.create_api()

        return api_id

    def find_resource_ids(self):
        """Given a resource path and API Id, find resource Id."""
        all_resources = self.client.get_resources(restApiId=self.api_id)
        parent_id = None
        resource_id = None
        for resource in all_resources['items']:
            if resource['path'] == "/":
                parent_id = resource['id']
            if resource['path'] == self.trigger_settings['resource']:
                resource_id = resource['id']
                self.log.info("Found Resource ID for: %s", resource['path'])
        return resource_id, parent_id

    def add_lambda_integration(self):
        """Attach lambda found to API."""
        lambda_uri = self.generate_uris()['lambda_uri']
        api_type = None
        if 'api_type' in self.trigger_settings:
            api_type = self.trigger_settings['api_type']
            self.log.info("Found API Integration Type: %s", api_type)
        else:
            api_type = 'AWS'
        self.client.put_integration(
            restApiId=self.api_id,
            resourceId=self.resource_id,
            httpMethod=self.trigger_settings['method'],
            integrationHttpMethod='POST',
            uri=lambda_uri,
            type=api_type)
        self.add_integration_response()
        self.log.info("Successfully added Lambda intergration to API")

    def add_integration_response(self):
        """Add an intergation response to the API for the lambda integration."""
        self.client.put_integration_response(
            restApiId=self.api_id,
            resourceId=self.resource_id,
            httpMethod=self.trigger_settings['method'],
            statusCode='200',
            responseTemplates={'application/json': ''})

    def add_permission(self):
        """Add permission to Lambda for the API Trigger."""
        statement_id = '{}_api_{}'.format(self.app_name, self.trigger_settings['api_name'])
        principal = 'apigateway.amazonaws.com'
        lambda_alias_arn = get_lambda_alias_arn(self.app_name, self.env, self.region)
        lambda_unqualified_arn = get_lambda_arn(self.app_name, self.env, self.region)
        resource_name = self.trigger_settings.get('resource', '')
        resource_name = resource_name.replace('/', '')
        method_api_source_arn = 'arn:aws:execute-api:{}:{}:{}/{}/{}/{}'.format(
            self.region, self.account_id, self.api_id, self.env, self.trigger_settings['method'], resource_name)
        global_api_source_arn = 'arn:aws:execute-api:{}:{}:{}/*/*/{}'.format(self.region, self.account_id, self.api_id,
                                                                             resource_name)
        add_lambda_permissions(
            function=lambda_alias_arn,
            statement_id=statement_id + self.trigger_settings['method'],
            action='lambda:InvokeFunction',
            principal=principal,
            env=self.env,
            region=self.region,
            source_arn=method_api_source_arn)
        add_lambda_permissions(
            function=lambda_alias_arn,
            statement_id=statement_id,
            action='lambda:InvokeFunction',
            principal=principal,
            env=self.env,
            region=self.region,
            source_arn=global_api_source_arn)
        add_lambda_permissions(
            function=lambda_unqualified_arn,
            statement_id=statement_id + self.trigger_settings['method'],
            action='lambda:InvokeFunction',
            principal=principal,
            env=self.env,
            region=self.region,
            source_arn=method_api_source_arn)
        add_lambda_permissions(
            function=lambda_unqualified_arn,
            statement_id=statement_id,
            action='lambda:InvokeFunction',
            principal=principal,
            env=self.env,
            region=self.region,
            source_arn=global_api_source_arn)

    @retries(max_attempts=5, wait=2, exceptions=(botocore.exceptions.ClientError))
    def create_api_deployment(self):
        """Create API deployment of ENV name."""
        try:
            self.client.create_deployment(restApiId=self.api_id, stageName=self.env)
            self.log.info('Created a deployment resource.')
        except botocore.exceptions.ClientError as error:
            error_code = error.response['Error']['Code']
            if error_code == 'TooManyRequestsException':
                self.log.debug('Retrying. We have hit api limit.')
            else:
                self.log.debug('Retrying. We received %s.', error_code)

    def create_api_key(self):
        """Create API Key for API access."""
        apikeys = self.client.get_api_keys()
        for key in apikeys['items']:
            if key['name'] == self.app_name:
                self.log.info("Key %s already exists", self.app_name)
                break
        else:
            self.client.create_api_key(
                name=self.app_name, enabled=True, stageKeys=[{
                    'restApiId': self.api_id,
                    'stageName': self.env
                }])
            self.log.info("Successfully created API Key %s. Look in the AWS console for the key", self.app_name)

    def _format_base_path(self, api_name):
        """Format the base path name."""
        name = self.app_name
        if self.app_name != api_name:
            name = '{0}-{1}'.format(self.app_name, api_name)
        return name

    def update_api_mappings(self):
        """Create a cname for the API deployment."""
        response_provider = None
        response_action = None
        domain = self.generated.apigateway()['domain']
        try:
            response_provider = self.client.create_base_path_mapping(
                domainName=domain,
                basePath=self._format_base_path(self.trigger_settings['api_name']),
                restApiId=self.api_id,
                stage=self.env, )
            response_action = 'API mapping added.'
        except botocore.exceptions.ClientError as error:
            error_code = error.response['Error']['Code']
            if error_code == 'ConflictException':
                response_action = 'API mapping already exist.'
            else:
                response_action = 'Unknown error: {0}'.format(error_code)

        self.log.debug('Provider response: %s', response_provider)
        self.log.info(response_action)
        return response_provider

    def generate_uris(self):
        """Generate several lambda uris."""
        lambda_arn = "arn:aws:execute-api:{0}:{1}:{2}/*/{3}/{4}".format(self.region, self.account_id, self.api_id,
                                                                        self.trigger_settings['method'],
                                                                        self.trigger_settings['resource'])

        lambda_uri = ("arn:aws:apigateway:{0}:lambda:path/{1}/functions/"
                      "arn:aws:lambda:{0}:{2}:function:{3}/invocations").format(self.region, self.api_version,
                                                                                self.account_id, self.app_name)

        api_dns = "https://{0}.execute-api.{1}.amazonaws.com/{2}".format(self.api_id, self.region, self.env)

        uri_dict = {'lambda_arn': lambda_arn, 'lambda_uri': lambda_uri, 'api_dns': api_dns}
        return uri_dict

    def create_api(self):
        """Create the REST API."""
        created_api = self.client.create_rest_api(name=self.trigger_settings.get('api_name', self.app_name))
        api_id = created_api['id']
        self.log.info("Successfully created API")
        return api_id

    def create_resource(self, parent_id=""):
        """Create the specified resource.

        Args:
            parent_id (str): The resource ID of the parent resource in API Gateway
        """
        resource_name = self.trigger_settings.get('resource', '')
        resource_name = resource_name.replace('/', '')
        if not self.resource_id:
            created_resource = self.client.create_resource(
                restApiId=self.api_id, parentId=parent_id, pathPart=resource_name)
            self.resource_id = created_resource['id']
            self.log.info("Successfully created resource")
        else:
            self.log.info("Resource already exists. To update resource please delete existing resource: %s",
                          resource_name)

    def attach_method(self, resource_id):
        """Attach the defined method."""
        try:
            _response = self.client.put_method(
                restApiId=self.api_id,
                resourceId=resource_id,
                httpMethod=self.trigger_settings['method'],
                authorizationType="NONE",
                apiKeyRequired=False, )
            self.log.debug('Response for resource (%s) push authorization: %s', resource_id, _response)
            _response = self.client.put_method_response(
                restApiId=self.api_id,
                resourceId=resource_id,
                httpMethod=self.trigger_settings['method'],
                statusCode='200')
            self.log.debug('Response for resource (%s) no authorization: %s', resource_id, _response)

            self.log.info("Successfully attached method: %s", self.trigger_settings['method'])
        except botocore.exceptions.ClientError:
            self.log.info("Method %s already exists", self.trigger_settings['method'])

    def setup_lambda_api(self):
        """A wrapper for all the steps needed to setup the integration."""
        self.create_resource(self.parent_id)
        self.attach_method(self.resource_id)
        self.add_lambda_integration()
        self.add_permission()
        self.create_api_deployment()
        self.create_api_key()
        self.update_api_mappings()



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Create Pipelines for Spinnaker."""
import collections
import json
import logging
import os
from pprint import pformat

from ..consts import DEFAULT_RUN_AS_USER, EC2_PIPELINE_TYPES
from ..exceptions import SpinnakerPipelineCreationFailed
from ..utils import ami_lookup, generate_packer_filename, get_details, get_properties, get_subnets, get_template
from ..utils.gate import gate_request
from .clean_pipelines import clean_pipelines
from .construct_pipeline_block import construct_pipeline_block
from .renumerate_stages import renumerate_stages


class SpinnakerPipeline:
    """Manipulate Spinnaker Pipelines.

    Args:
        app (str): Application name.
        trigger_job (str): Jenkins trigger job.
        base (str): Base image name (i.e: fedora).
        prop_path (str): Path to the raw.properties.json.
        runway_dir (str): Path to local runway directory.
    """

    def __init__(self, app='', trigger_job='', prop_path='', base='', runway_dir=''):
        self.log = logging.getLogger(__name__)

        self.header = {'content-type': 'application/json'}
        self.here = os.path.dirname(os.path.realpath(__file__))

        self.runway_dir = os.path.expandvars(os.path.expanduser(runway_dir or ''))

        self.base = base
        self.trigger_job = trigger_job
        self.generated = get_details(app=app)
        self.app_name = self.generated.app_name()
        self.group_name = self.generated.project
        self.repo_name = self.generated.repo

        self.settings = get_properties(prop_path)
        self.environments = self.settings['pipeline']['env']

    def post_pipeline(self, pipeline):
        """Send Pipeline JSON to Spinnaker.

        Args:
            pipeline (json): json of the pipeline to be created in Spinnaker
        """
        uri = '/pipelines'

        if isinstance(pipeline, str):
            pipeline_json = pipeline
        else:
            pipeline_json = json.dumps(pipeline)

        pipeline_dict = json.loads(pipeline_json)

        self.log.debug('Pipeline JSON:\n%s', pipeline_json)

        pipeline_response = gate_request(method='POST', uri=uri, data=pipeline_json, headers=self.header)

        self.log.debug('Pipeline creation response:\n%s', pipeline_response.text)

        if not pipeline_response.ok:
            raise SpinnakerPipelineCreationFailed('Pipeline for {0}: {1}'.format(self.app_name,
                                                                                 pipeline_response.json()))

        self.log.info('Successfully created "%s" pipeline in application "%s".', pipeline_dict['name'],
                      pipeline_dict['application'])

    def render_wrapper(self, region='us-east-1'):
        """Generate the base Pipeline wrapper.

        This renders the non-repeatable stages in a pipeline, like jenkins, baking, tagging and notifications.

        Args:
            region (str): AWS Region.

        Returns:
            dict: Rendered Pipeline wrapper.

        """
        base = self.settings['pipeline']['base']

        if self.base:
            base = self.base

        email = self.settings['pipeline']['notifications']['email']
        slack = self.settings['pipeline']['notifications']['slack']
        baking_process = self.settings['pipeline']['image']['builder']
        provider = 'aws'
        root_volume_size = self.settings['pipeline']['image']['root_volume_size']
        bake_instance_type = self.settings['pipeline']['image']['bake_instance_type']

        ami_id = ami_lookup(name=base, region=region)

        ami_template_file = generate_packer_filename(provider, region, baking_process)

        pipeline_id = self.compare_with_existing(region=region)

        data = {
            'app': {
                'ami_id': ami_id,
                'appname': self.app_name,
                'group_name': self.group_name,
                'repo_name': self.repo_name,
                'base': base,
                'environment': 'packaging',
                'region': region,
                'triggerjob': self.trigger_job,
                'run_as_user': DEFAULT_RUN_AS_USER,
                'email': email,
                'slack': slack,
                'root_volume_size': root_volume_size,
                'bake_instance_type': bake_instance_type,
                'ami_template_file': ami_template_file,
                'pipeline': self.settings['pipeline']
            },
            'id': pipeline_id
        }

        self.log.debug('Wrapper app data:\n%s', pformat(data))

        wrapper = get_template(template_file='pipeline/pipeline_wrapper.json.j2', data=data, formats=self.generated)

        return json.loads(wrapper)

    def get_existing_pipelines(self):
        """Get existing pipeline configs for specific application.

        Returns:
            str: Pipeline config json

        """
        uri = "/applications/{0}/pipelineConfigs".format(self.app_name)
        resp = gate_request(uri=uri)
        assert resp.ok, 'Failed to lookup pipelines for {0}: {1}'.format(self.app_name, resp.text)

        return resp.json()

    def compare_with_existing(self, region='us-east-1', onetime=False):
        """Compare desired pipeline with existing pipelines.

        Args:
            region (str): Region of desired pipeline.
            onetime (bool): Looks for different pipeline if Onetime

        Returns:
            str: pipeline_id if existing, empty string of not.

        """
        pipelines = self.get_existing_pipelines()
        pipeline_id = None
        found = False
        for pipeline in pipelines:
            correct_app_and_region = (pipeline['application'] == self.app_name) and (region in pipeline['name'])
            if onetime:
                onetime_str = "(onetime-{})".format(self.environments[0])
                if correct_app_and_region and onetime_str in pipeline['name']:
                    found = True
            elif correct_app_and_region:
                found = True

            if found:
                self.log.info('Existing pipeline found - %s', pipeline['name'])
                pipeline_id = pipeline['id']
                break
        else:
            self.log.info('No existing pipeline found')

        return pipeline_id

    def create_pipeline(self):
        """Main wrapper for pipeline creation.
        1. Runs clean_pipelines to clean up existing ones
        2. determines which environments the pipeline needs
        3. gets all subnets for template rendering
        4. Renders all of the pipeline blocks as defined in configs
        5. Runs post_pipeline to create pipeline
        """
        clean_pipelines(app=self.app_name, settings=self.settings)

        pipeline_envs = self.environments
        self.log.debug('Envs from pipeline.json: %s', pipeline_envs)

        regions_envs = collections.defaultdict(list)
        for env in pipeline_envs:
            for region in self.settings[env]['regions']:
                regions_envs[region].append(env)
        self.log.info('Environments and Regions for Pipelines:\n%s', json.dumps(regions_envs, indent=4))

        subnets = None
        pipelines = {}
        for region, envs in regions_envs.items():
            self.generated.data.update({
                'region': region,
            })

            # TODO: Overrides for an environment no longer makes sense. Need to
            # provide override for entire Region possibly.
            pipelines[region] = self.render_wrapper(region=region)

            previous_env = None
            for env in envs:
                self.generated.data.update({
                    'env': env,
                })

                pipeline_block_data = {
                    "env": env,
                    "generated": self.generated,
                    "previous_env": previous_env,
                    "region": region,
                    "settings": self.settings[env][region],
                    "pipeline_data": self.settings['pipeline'],
                }

                if self.settings['pipeline']['type'] in EC2_PIPELINE_TYPES:
                    if not subnets:
                        subnets = get_subnets()
                    try:
                        region_subnets = {region: subnets[env][region]}
                    except KeyError:
                        self.log.info('%s is not available for %s.', env, region)
                        continue
                    pipeline_block_data['region_subnets'] = region_subnets

                block = construct_pipeline_block(**pipeline_block_data)
                pipelines[region]['stages'].extend(json.loads(block))
                previous_env = env

        self.log.debug('Assembled Pipelines:\n%s', pformat(pipelines))

        for region, pipeline in pipelines.items():
            renumerate_stages(pipeline)

            self.post_pipeline(pipeline)

        return True



#   Foremast - Pipeline Tooling
#
#   Copyright 2020 Redbox Automated Retail, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Client for deploying GCP Cloud Functions"""

import logging
import requests
from googleapiclient.errors import HttpError
from googleapiclient import discovery
from ..exceptions import CloudFunctionDeployError, CloudFunctionOperationFailedError, \
    CloudFunctionOperationIncompleteError
from tryagain import retries
from ..utils.gcp_environment import GcpEnvironment

LOG = logging.getLogger(__name__)


class CloudFunctionsClient:

    def __init__(self, app_name, env: GcpEnvironment, configs, use_service_account=True):
        """Creates a new GcpEnvironment object
        Args:
            app_name (str): Full name off app (e.g. coolappneatproject)
            env (GcpEnvironment): GCP Environment to target
            configs (dict): Pipeline configs
            use_service_account (bool): When true a service account {app_name}@{project}.iam.gserviceaccount.com is
                assigned to the function.  It is assumed this service account exists already.  Defaults to True.
        """
        credentials = env.get_credentials()
        api_builder = discovery.build('cloudfunctions', 'v1', credentials=credentials)
        self._env = env
        self._functions_client = api_builder.projects().locations().functions()
        self._operations_client = api_builder.operations()
        self._pipeline_config = configs['pipeline']  # pipeline.json options
        self._cf_config = self._pipeline_config['cloudfunction']  # pipeline.json, cloudfunction block
        self._env_config = configs[env.name]  # application-master-{env}.json options
        self._name = app_name
        self._project_id = None
        self._regions = list()
        self._use_service_account = use_service_account
        self._service_account_email = None

    def prepare_client(self):
        # Load project name, this may require API calls to GCP
        # and should not be in __init__
        try:
            project_wildcard = self._cf_config["project_name"]
            self._project_id = self._env.get_project(project_wildcard)['projectId']
        except KeyError:
            LOG.error("Missing required field: cloudfunction.project_name in pipeline.json")
            raise

        try:
            self._regions = self._env_config['regions']
            if not self._regions:
                raise ValueError("At least one region must be defined")
        except (KeyError, ValueError):
            LOG.error("Cloud Function pipelines require at least one region in pipeline.json")
            raise

        if self._use_service_account:
            self._service_account_email = "{}@{}.iam.gserviceaccount.com".format(self._name, self._project_id)
            LOG.info("Using dedicated service account '%s'", self._service_account_email)

    def deploy_function(self, file_path, region):
        """Creates (or updates) and uploads a function's code in one step

        Args:
            file_path: str, required
                Local file path to the function's zip file
            region: str, required
                The region to deploy to

        Returns:
            None

        Raises:
            CloudFunctionOperationIncompleteError: Timeout waiting for the operation to complete
            CloudFunctionOperationFailedError: Operation completed with error
            HttpError: Error communicating with the GCP APIs
        """
        if region not in self._regions:
            raise CloudFunctionDeployError("Region '{}' is not in cloud function's configuration.  Options are: {}"
                                           .format(region, self._regions))

        upload_url = self._get_upload_url(region)
        self._upload_zip(upload_url, file_path)
        exists = self._check_function_exists(region)
        # Deploy function, then update it's IAM Policy for access
        if exists:
            self._update_function(region, upload_url)
        else:
            self._create_function(region, upload_url)
        self._update_function_iam_policy(region)

    def _get_upload_url(self, location_id):
        """Gets a Signed URL that the function's zip file can be uploaded to

        Args:
            location_id: str, required
                Location Id (region) the function will be uploaded to.  (e.g. us-east1)
        """
        parent = "projects/{}/locations/{}".format(self._project_id, location_id)
        result = self._functions_client.generateUploadUrl(parent=parent, body={}).execute()
        return result["uploadUrl"]

    def _upload_zip(self, upload_url, file_path):
        """Uploads a function's local zip file to the destination url given

        Args:
            upload_url: str, required
                The signed URL generated using get_upload_url()
            file_path: str, required
                The local path to the zip file to be uploaded
        """

        with open(file_path, 'rb') as file:
            # Note: The upload_url will be a signed URL generated for us
            # Do not set BEARER auth header
            headers = {
                'content-type': 'application/zip',
                'x-goog-content-length-range': '0,104857600'
            }
            result = requests.put(upload_url, data=file, headers=headers)
            result.raise_for_status()
            LOG.info("Successfully uploaded cloud function code")

    def _create_function(self, location_id, upload_url):
        """Creates a new Cloud Function

            Args:
                location_id (str): Location Id (Region)

            Returns:
                None

            Raises:
                CloudFunctionOperationIncompleteError: Timeout waiting for the operation to complete
                CloudFunctionOperationFailedError: Operation completed with error
                HttpError: Error communicating with the GCP APIs
        """
        parent = "projects/{}/locations/{}".format(self._project_id, location_id)
        request_body = self._generate_function_request_body(location_id, upload_url)

        LOG.info("Creating cloud function '%s' in region '%s' and project '%s'",
                 self._name, location_id, self._project_id)
        response = self._functions_client.create(location=parent, body=request_body).execute()
        # If operation times out or fails an exception will be raised to stop the pipeline
        self._wait_for_operation(response['name'])
        LOG.info("Successfully created cloud function '%s' in region '%s' and project '%s'",
                 self._name, location_id, self._project_id)

    def _update_function(self, location_id, upload_url):
        """Updates an existing Cloud Function

            Args:
                location_id (str): Location Id (Region)

            Returns:
                None

            Raises:
                CloudFunctionOperationIncompleteError: Timeout waiting for the operation to complete
                CloudFunctionOperationFailedError: Operation completed with error
                HttpError: Error communicating with the GCP APIs
        """
        full_name = self._generate_function_path(location_id)
        request_body = self._generate_function_request_body(location_id, upload_url)
        LOG.info("Updating cloud function '%s' in region '%s' and project '%s'",
                 self._name, location_id, self._project_id)
        response = self._functions_client.patch(name=full_name, body=request_body).execute()
        # If operation times out or fails an exception will be raised to stop the pipeline
        self._wait_for_operation(response['name'])
        LOG.info("Successfully updated cloud function '%s' in region '%s' and project '%s'",
                 self._name, location_id, self._project_id)

    def _update_function_iam_policy(self, region: str):
        """Updates a function's individual IAM Policy which is used to control access to the cloud function
        For example, limiting access to a certain service account or allowing unauthenticated access

            Args:
                region (str): Region the function is deployed to (e.g. us-east1)

            Returns:
                None
        """
        resource_name = self._generate_function_path(region)
        bindings = []
        # If they use the helper allow_unauthenticated add the binding for them
        # Equivilent of gcloud functions deploy ... --allow-unauthenticated
        allow_unauthenticated = self._env_config['app'].get("cloudfunction_allow_unauthenticated", False)
        custom_bindings = self._env_config['app'].get("cloudfunction_iam_bindings", [])
        if allow_unauthenticated:
            bindings.append(CloudFunctionsClient._get_allow_unauthenticated_binding())
        if custom_bindings:
            bindings.extend(custom_bindings)
        LOG.info("Updating Cloud Function IAM Policy for '%s' in region '%s' and project '%s': Bindings '%s'",
                 self._name, region, self._project_id, bindings)
        body_payload = {
            "policy": {
                "bindings": bindings
            }
        }
        self._functions_client.setIamPolicy(resource=resource_name, body=body_payload).execute()

    def _check_function_exists(self, location_id):
        """Checks a GCP Cloud Function exists

            Args:
                location_id (str): Region to check for the function in

            Returns:
                True: When function exists
                False: When function does not exist
        """
        full_name = self._generate_function_path(location_id)
        try:
            self._functions_client.get(name=full_name).execute()
            return True
        except HttpError as e:
            # 404 is expected if the function does not exist
            # any other HttpError would should be re-raised
            if 'status' in e.resp and e.resp['status'] == '404':
                return False
            raise e

    # Retry with back off, wait time is attempt_count * 2 (1 second, 2 seconds, 4 seconds, etc.)
    @retries(max_attempts=10, wait=lambda n: 2 ** n, exceptions=CloudFunctionOperationIncompleteError)
    def _wait_for_operation(self, operation_name):
        """Waits for the given operation to complete with an exponential back off.  The back off is
        the number of attempts multiplied by two (in seconds).  Maximum 10 attempts total, for a total of
        110 seconds waiting before timing out.

            Args:
                operation_name (str): Name of the operation in form 'operation/{id}'

            Raises:
                CloudFunctionOperationIncompleteError: Timeout waiting for the operation to complete
                CloudFunctionOperationFailedError: Operation completed with error
            """
        LOG.info("Waiting for operation to complete.  Will poll GCP periodically.")
        response = self._operations_client.get(name=operation_name).execute()
        # If done is false, or no 'done' key is present, the operation is still running
        # If done, and error are set, it completed with a failure.  Typically this means there is
        # an issue with the function's code and the user should be made aware
        done = response.get('done', False)
        error = response.get('error', None)
        if done and error:
            raise CloudFunctionOperationFailedError(error, "Cloud Function deployment failed: {}".format(error))
        elif done:
            return
        else:
            # Raise error to trigger back off
            raise CloudFunctionOperationIncompleteError()

    def _generate_function_path(self, location_id):
        return "projects/{}/locations/{}/functions/{}".format(self._project_id, location_id, self._name)

    def _generate_function_request_body(self, location_id, upload_url):
        """Generates the request body for create and update/patch cloud function API calls to GCP using
        defaults and pipeline configuration files for the current repo.

           Args:
               location_id (str): The region being deployed to
               upload_url (str): The URL of the already uploaded/zipped source code

           Returns:
               dict: Response body for GCP create/patch function API calls
       """
        app_config = self._env_config['app']
        vpc_connector_block = app_config.get("cloudfunction_vpc", {})

        request_body = {
            # Automated options
            "name": self._generate_function_path(location_id),
            "serviceAccountEmail": self._service_account_email,
            "sourceUploadUrl": upload_url,
            # General pipeline options
            "entryPoint": self._cf_config['entry_point'],
            "runtime": self._cf_config['runtime'],
            "labels": {},
            # Env specific options
            "timeout": app_config.get("cloudfunction_timeout"),
            "availableMemoryMb": app_config.get("cloudfunction_memory_mb"),
            "environmentVariables": app_config.get("cloudfunction_environment"),
            "maxInstances": app_config.get("cloudfunction_max_instances"),
            # Get the connector name for this region
            "vpcConnector":  vpc_connector_block.get("connector", {}).get(location_id),
            "vpcConnectorEgressSettings": vpc_connector_block.get("egress_type"),
            "ingressSettings": app_config.get("cloudfunction_ingress_type")
        }

        # GCP only supports either an HTTP trigger or an event trigger, both can not be used
        # HTTP Trigger is the default and can work with no configuration
        # if they did not give an event trigger, assume HTTP
        event_trigger = app_config.get("cloudfunction_event_trigger")
        if event_trigger:
            if "failure_policy" in event_trigger and event_trigger["failure_policy"].get("retry"):
                # GCP API specifies this as an object, but it is more of a bool depending on if the obj is set or not
                retry = {}
            else:
                retry = None
            request_body["eventTrigger"] = {
                "eventType": event_trigger["event_type"],
                "resource": self._get_full_resource_path(event_trigger["resource"]),
                "service": event_trigger.get("service"),
                "failurePolicy": {
                    "retry": retry
                }
            }
        else:
            request_body["httpsTrigger"] = {}

        return request_body

    def _get_full_resource_path(self, partial_path):
        """Translates a partial resource path `topics/my_topic` or `/topics/my_topic`
        to the full path including project, like `/projects/my-project/topics/my_topic`.
        Most GCP APIs expect the full path.  The deployment target project is used, as GCP
        Functions can only be triggered by a resource in the same project.

        Args:
            partial_path (str): The partial resource path

        GCP Docs: https://cloud.google.com/functions/docs/reference/rest/v1/projects.locations.functions#EventTrigger"""

        if partial_path.startswith("projects"):
            raise CloudFunctionDeployError("Path '{}' contains a hardcoded project identifier. ".format(partial_path) +
                                           "Foremast adds this automatically based on the target deployment project. " +
                                           "Path should start at the event type (e.g. databases, bucket, etc).")

        return "projects/{}/{}".format(self._project_id, partial_path.lstrip('/'))

    @staticmethod
    def _get_allow_unauthenticated_binding():
        """
        Gets an IAM Policy binding that will allow anonymous access to a Cloud Function
        See: https://cloud.google.com/functions/docs/securing/managing-access-iam
        Is equivilent of gcloud function deploy ... --allow-unauthenticated

        Returns:
            Dict, Single IAM Policy binding object: https://cloud.google.com/functions/docs/reference/rest/v1/Policy
        """
        return {
            "members": [
                "allUsers"
            ],
            "role": "roles/cloudfunctions.invoker"
        }



#   Foremast - Pipeline Tooling
#
#   Copyright 2019 Redbox Automated Retail, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Manages AWS ASG Scheduled Actions in Spinnaker. Can find, create, and delete.
"""
import logging
import os

from ..utils import get_latest_server_group, get_properties, get_template, wait_for_task


class ScheduledActions:
    """Manages scheduled actions in Spinnaker

    Args:
        app (str): Application name
        prop_path (str): Path of rendered property files
        env (str): Environment/Account to add scheduled actions to
        region (str): AWS region for scheduled actions

    Attributes:
        log (str): Logger name
        settings (dict): Properties imported from prop_path
    """

    def __init__(self, app='', prop_path='', env='', region=''):

        self.log = logging.getLogger(__name__)

        self.here = os.path.dirname(os.path.realpath(__file__))
        self.env = env
        self.region = region
        self.app = app

        self.settings = get_properties(properties_file=prop_path, env=self.env, region=self.region)

    def prepare_scheduled_actions_template(self, scheduled_actions, server_group):
        """Renders scheduled actions templates based on configs and variables.
        After rendering, POSTs the json to Spinnaker for creation.

        Args:
            server_group (str): The name of the server group to render template for
        """
        template_kwargs = {
            'app': self.app,
            'env': self.env,
            'region': self.region,
            'server_group': server_group,
            'scheduled_actions': scheduled_actions
        }

        rendered_template = get_template(template_file='infrastructure/scheduled_actions.json.j2', **template_kwargs)
        self.log.info('Creating scheduled actions in %s for %s', self.env, self.app)
        wait_for_task(rendered_template)
        self.log.info('Successfully created scheduled actions in %s for %s', self.env, self.app)

    def create_scheduled_actions(self):
        """Wrapper function. Gets the latest server group and then runs
        self.prepare_scheduled_actions_template for scheduled actions. This
        function acts as the main driver for the scheduled actions creation.
        """
        if not self.settings['asg']['scheduled_actions']:
            self.log.info("No scheduled actions found, skipping...")
            return

        server_group = get_latest_server_group(self.env, self.app)
        scheduled_actions = self.settings['asg']['scheduled_actions']
        self.prepare_scheduled_actions_template(scheduled_actions, server_group)



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Test Git file lookups."""
import base64
from unittest import mock

import pytest

from foremast.exceptions import GitLabApiError
from foremast.utils import FileLookup

TEST_JSON = '''{
    "ship": "pirate"
}'''
TEST_JSON_BYTES = TEST_JSON.encode()


@mock.patch('foremast.utils.lookups.gitlab')
def test_init(gitlab):
    """Check init."""
    my_git = FileLookup()
    assert my_git.git_short == ''
    assert my_git.server == gitlab.Gitlab.return_value
    my_git.server.projects.get.assert_called_with(my_git.git_short)


@mock.patch('foremast.utils.lookups.gitlab')
def test_project_exception(mock_gitlab):
    """Check resolving GitLab Project ID fails with exception."""
    mock_gitlab.Gitlab.return_value.projects.get.return_value = False

    with pytest.raises(GitLabApiError):
        FileLookup()


@mock.patch('foremast.utils.lookups.gitlab')
def test_project_success(mock_gitlab):
    """Check resolving GitLab Project ID is successful."""
    mock_gitlab.Gitlab.return_value.projects.get.return_value = object
    seeker = FileLookup()
    assert seeker.project is object


@mock.patch.object(FileLookup, 'remote_file', return_value=TEST_JSON)
@mock.patch('foremast.utils.lookups.gitlab')
def test_get(gitlab, mock_lookup):
    """Check _get_ method."""
    my_git = FileLookup()

    result = my_git.get()
    assert isinstance(result, str)
    assert TEST_JSON == my_git.get()


@mock.patch.object(FileLookup, 'get', return_value=TEST_JSON)
@mock.patch('foremast.utils.lookups.gitlab')
def test_json(gitlab, mock_lookup):
    """Check _json_ method."""
    my_git = FileLookup()

    result = my_git.json()
    assert isinstance(result, dict)
    assert result['ship'] == 'pirate'


@mock.patch.object(FileLookup, 'get', return_value=TEST_JSON + '}')
@mock.patch('foremast.utils.lookups.gitlab')
def test_invalid_json(gitlab, mock_lookup):
    """Check invalid JSON."""
    my_git = FileLookup()

    with pytest.raises(SystemExit):
        my_git.json()


def test_filelookup_attr():
    """Make sure Git related attributes are missing when runway is specified."""
    my_git = FileLookup(runway_dir='/poop_deck')

    assert my_git.git_short == ''
    assert my_git.server is None


@mock.patch('foremast.utils.lookups.gitlab')
def test_runway_get(gitlab, tmpdir):
    """Make sure Git is not called when runway is specified."""
    filename = 'test.json'

    tmpdir.join(filename).write(TEST_JSON)

    my_git = FileLookup(runway_dir=str(tmpdir))
    result = my_git.get(filename=filename)

    assert isinstance(result, str)
    assert result == TEST_JSON

    with pytest.raises(FileNotFoundError):
        my_git.get(filename='parrot')
